#####################################################################################
# A tool for the creation of JasperGold SVP principle tcl files.
# Copyright (C) 2024  RISCY-Lib Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#####################################################################################

from __future__ import annotations
from typing import Callable, override

import contextlib

from PySide6.QtWidgets import (
    QMainWindow, QTableWidget, QDockWidget, QWidget, QFormLayout,
    QTextEdit, QLineEdit, QComboBox, QTableWidgetItem, QCheckBox
)
from PySide6.QtCore import Qt

from mavsec import properties
from mavsec.project import Project, ProjectInfo
from mavsec.properties import Property


####################################################################################################
# Project Tab
####################################################################################################
class ProjectTab(QTableWidget):
    """The project information pane"""
    def __init__(
                    self,
                    parent: QWidget,
                    proj: Project | None = None,
                    pdock: PropertyDock | None = None
                ):
        super().__init__(parent)
        self.setObjectName(u"properties_table")

        self._pdock = pdock

        columns = ["Name", "Type", "Description"]
        self.setColumnCount(len(columns))
        self.setColumnWidth(0, 250)
        self.setColumnWidth(1, 250)
        self.setColumnWidth(2, 500)
        self.setHorizontalHeaderLabels(columns)
        self.setEditTriggers(QTableWidget.EditTrigger.NoEditTriggers)

        if proj is None:
            self._proj = Project(ProjectInfo("", "", ""))
        else:
            self._proj = proj

        for i in range(len(self._proj.properties)):
            super().insertRow(i)
            self.set_row(i)

    def activate(self) -> None:
        self.cellPressed.connect(self._cell_pressed)

    def _cell_pressed(self, row: int, col: int) -> None:
        if self._pdock is not None:
            self._pdock.activate(self._proj.properties, row, self.update_current_row)

    def deactivate(self) -> None:
        with contextlib.suppress(RuntimeError):
            self.cellPressed.disconnect()

        if self._pdock is not None:
            self._pdock.deactivate()

    @override
    def insertRow(self, row: int) -> None:  # noqa: N802 this is a parent class method
        self._proj.properties.insert(row, Property("", "", ptype=properties.SecureKeyProperty))
        super().insertRow(row)
        self.set_row(row)

    def set_row(self, row: int) -> None:
        self.setItem(row, 0, QTableWidgetItem(self._proj.properties[row].name))
        self.setItem(row, 1, QTableWidgetItem(self._proj.properties[row].type_name()))
        self.setItem(row, 2, QTableWidgetItem(self._proj.properties[row].description))

    def update_current_row(self) -> None:
        row = self.currentRow()
        self.set_row(row)

    def get_proj(self) -> Project:
        return self._proj


####################################################################################################
# Property Information Pane
####################################################################################################
class PropertyDock(QDockWidget):
    """The property information pane"""
    def __init__(self, parent: QMainWindow):
        super().__init__("Property Info", parent=parent)
        self.setFeatures(
            QDockWidget.DockWidgetFeature.DockWidgetMovable |
            QDockWidget.DockWidgetFeature.DockWidgetFloatable
        )
        parent.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self)

        self._form = QWidget(self)
        self._layout = QFormLayout(self._form)
        self._form.setLayout(self._layout)

        self._name = QLineEdit(self._form)
        self._type = QComboBox(self._form)
        self._description = QTextEdit(self._form)
        self._preconditions = QTextEdit(self._form)

        self._type.addItems([t.name for t in Property.available_types()])

        self._layout.addRow("Name", self._name)
        self._layout.addRow("Type", self._type)
        self._layout.addRow("Description", self._description)
        self._layout.addRow("Preconditions", self._preconditions)

        self.setDisabled(True)

        self.setWidget(self._form)

    def activate(self, props: list[Property], row: int, update: Callable) -> None:
        self.deactivate()

        self._props = props
        self._prow = row
        self._update = update
        prop = props[row]
        self._name.setText(prop.name)
        self._type.setCurrentText(prop.type_name())
        self.set_type(prop.type_name())
        self._description.setText(prop.description)

        self._name.textChanged.connect(lambda text: setattr(prop, "name", text))
        self._name.textChanged.connect(update)
        self._type.currentTextChanged.connect(self.set_type)
        self._type.currentTextChanged.connect(update)
        self._description.textChanged.connect(
            lambda: setattr(prop, "description", self._description.toPlainText())
        )
        self._description.textChanged.connect(update)
        self._preconditions.textChanged.connect(
            lambda: setattr(prop, "preconditions", self._preconditions.toPlainText())
        )

        self.setDisabled(False)

    def set_type(self, text: str) -> None:
        ptype = properties.Property.ptype_from_str(text)
        self._props[self._prow].ptype = ptype

        self.remove_type_fields()
        self.add_type_fields()

    def deactivate(self) -> None:
        with contextlib.suppress(RuntimeError):
            self._name.textChanged.disconnect()
            self._description.textChanged.disconnect()

        self._name.setText("")
        self._description.setText("")

        self.remove_type_fields()

        self.setDisabled(True)

    def remove_type_fields(self) -> None:
        rows = self._layout.rowCount()
        for i in range(rows-1, 3, -1):
            self._layout.removeRow(i)

    def add_type_fields(self) -> None:
        prop = self._props[self._prow]

        if prop.ptype is None:
            return
        elif isinstance(prop.ptype, str):
            prop.ptype = properties.PropertyType.get_type(prop.ptype)

        for meta, mtype in prop.ptype.meta.items():
            if mtype is bool:
                cb = QCheckBox(self._form)
                if meta not in prop.meta:
                    prop.meta[meta] = False
                cb.setChecked(prop.meta[meta])
                self._layout.addRow(meta, cb)
                cb.stateChanged.connect(lambda state, m=meta: setattr(prop.meta, m, bool(state)))
            elif mtype is str:
                le = QLineEdit(self._form)
                if meta not in prop.meta:
                    prop.meta[meta] = ""
                le.setText(prop.meta[meta])
                le.textChanged.connect(lambda text, m=meta: setattr(prop.meta, m, text))
                le.textChanged.connect(self._update)
                self._layout.addRow(meta, le)
            elif mtype is int:
                le = QLineEdit(self._form)
                if meta not in prop.meta:
                    prop.meta[meta] = 0
                le.setText(str(prop.meta[meta]))
                le.textChanged.connect(lambda text, m=meta: setattr(prop.meta, m, int(text)))
                le.textChanged.connect(self._update)
                self._layout.addRow(meta, le)
            elif mtype is properties.AnyRtlPath:
                le = QLineEdit(self._form)
                if meta not in prop.meta:
                    prop.meta[meta] = ""
                le.setText(prop.meta[meta])
                le.textChanged.connect(lambda text, m=meta: setattr(prop.meta, m, text))
                le.textChanged.connect(self._update)
                le.setPlaceholderText("Path to signal in RTL")
                self._layout.addRow(meta, le)


####################################################################################################
# Project Information Pane
####################################################################################################
def _project_setter(proj: Project, proj_attr: str) -> Callable[[str], None]:
    def _setter(text: str) -> None:
        setattr(proj.info, proj_attr, text)
    return _setter


class ProjectInfoDock(QDockWidget):

    _proj_ptr: Project | None

    """The project information pane"""
    def __init__(self, parent: QMainWindow):
        super().__init__("Project Info", parent=parent)
        self.setFeatures(
            QDockWidget.DockWidgetFeature.DockWidgetMovable |
            QDockWidget.DockWidgetFeature.DockWidgetFloatable
        )
        parent.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, self)

        form = QWidget(self)
        layout = QFormLayout(form)
        form.setLayout(layout)

        self._name = QLineEdit(form)
        self._version = QLineEdit(form)
        self._description = QTextEdit(form)

        layout.addRow("Project Name", self._name)
        layout.addRow("Project Version", self._version)
        layout.addRow("Project Description", self._description)

        self.setDisabled(True)

        self._proj_ptr = None

        self.setWidget(form)

    def clear(self) -> None:
        self._proj_ptr = None

        self._name.setText("")
        self._version.setText("")
        self._description.setText("")

    def set_proj(self, proj: Project) -> None:
        self._proj_ptr = proj
        self.set_from_proj(proj)

        with contextlib.suppress(RuntimeError):
            self._name.textChanged.disconnect()
            self._version.textChanged.disconnect()
            self._description.textChanged.disconnect()

        self._name.textChanged.connect(_project_setter(self._proj_ptr, "name"))
        self._version.textChanged.connect(_project_setter(self._proj_ptr, "version"))
        self._description.textChanged.connect(
            lambda: _project_setter(self._proj_ptr, "description")(self._description.toPlainText())
        )

    def set_from_proj(self, proj: Project) -> None:
        self._name.setText(proj.info.name)
        self._version.setText(proj.info.version)
        self._description.setText(proj.info.description)



#####################################################################################
# A tool for the creation of JasperGold SVP principle tcl files.
# Copyright (C) 2024  RISCY-Lib Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#####################################################################################

from __future__ import annotations
import pathlib

from PySide6.QtWidgets import (
  QMainWindow, QTabWidget, QFileDialog, QToolBar, QDialog, QLabel, QVBoxLayout, QListWidget
)
from PySide6 import QtGui, QtCore

import webbrowser

from mavsec.general import ProjectCheckResult
from mavsec.project import Project

from mavsec._gui.proj_tabs import ProjectTab, PropertyDock, ProjectInfoDock
from mavsec._gui.menus import MenuBar, StatusBar

FILE_FILTERS = [
    "Project (*.yaml *.yml *.toml *.json)",
    "YAML (*.yaml *.yml)",
    "TOML (*.toml)",
    "JSON (*.json)"
]


################################################################################################
# Main Window
################################################################################################
class MavSecMainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setObjectName(u"MavSecMainWindow")
        self.setWindowTitle("MavSec")

        self.resize(1920, 1080)

        self._statusBar = StatusBar(self)
        self.setStatusBar(self._statusBar)

        self._menuBar = MenuBar(self)
        self.setMenuBar(self._menuBar)
        self._setup_menu_bar()

        self._create_tabs()
        self._setup_tabs()

        self._create_toolbar()
        self._setup_toolbar()

        self._project = ProjectInfoDock(self)

        self._properties = PropertyDock(self)

    # Menu Bar Actions
    ################################################################################################
    def _setup_menu_bar(self) -> None:
        self._menuBar.file.actionNew.triggered.connect(self._new_project)
        self._menuBar.file.actionOpen.triggered.connect(self._open_project)
        self._menuBar.file.actionSave.triggered.connect(self._save_project)
        self._menuBar.file.actionExport.triggered.connect(self._export_project)
        self._menuBar.file.actionClose.triggered.connect(self._close_project)
        self._menuBar.file.actionQuit.triggered.connect(self._quit)

        self._menuBar.edit.actionNewProperty.triggered.connect(self._new_property_action)
        self._menuBar.edit.actionRemoveProperty.triggered.connect(self._remove_property_action)
        self._menuBar.edit.actionCheckSecurity.triggered.connect(self._check_security_action)
        self._menuBar.edit.actionCheckDesign.triggered.connect(self._check_implementation_action)
        self._menuBar.edit.actionCheckAll.triggered.connect(self._check_properties_action)

        self._menuBar.help.actionDocumentation.triggered.connect(self._open_documentation)

    def _new_project(self) -> None:
        tab_idx = self._tabs.addTab(ProjectTab(self._tabs, None, self._properties), "New Project")
        self._tabs.setCurrentIndex(tab_idx)

    def _open_project(self) -> None:
        filename, ok = QFileDialog.getOpenFileName(
            self,
            "Open Project",
            ".",
            ";;".join(FILE_FILTERS),
        )
        if not ok:
            return

        proj = Project.from_file(filename)
        proj.info.proj_file = filename

        proj_tab = ProjectTab(self._tabs, proj, self._properties)
        tab_idx = self._tabs.addTab(proj_tab, proj.info.name)
        self._tabs.setCurrentIndex(tab_idx)

    def _save_project(self) -> None:
        tab = self._tabs.currentWidget()
        if not isinstance(tab, ProjectTab):
            return

        self.save_tab(tab)

        t_idx = self._tabs.currentIndex()
        self._tabs.setTabText(t_idx, tab._proj.info.name)

    def _save_as_project(self) -> None:
        tab = self._tabs.currentWidget()
        if not isinstance(tab, ProjectTab):
            return

        self.save_tab_as(tab)

    def _save_all_projects(self) -> None:
        for tab_idx in range(self._tabs.count()):
            tab = self._tabs.widget(tab_idx)
            if not isinstance(tab, ProjectTab):
                continue

            self.save_tab(tab)

    def _export_project(self) -> None:
        tab = self._tabs.currentWidget()
        if not isinstance(tab, ProjectTab):
            return

        filename, ok = QFileDialog.getSaveFileName(
            self,
            "Export Project",
            ".",
            "TCL (*.tcl)",
        )

        if not ok:
            return

        tab.get_proj().to_tcl(filename)

    def _close_project(self) -> None:
        self._tabs.removeTab(self._tabs.currentIndex())

    def _quit(self) -> None:
        self._save_all_projects()

        self.close()

    def _check_security_action(self) -> None:
        tab = self._tabs.currentWidget()
        if not isinstance(tab, ProjectTab):
            return

        errors = tab.get_proj().check()
        errors.design_errors = []
        errors.design_warnings = []

        if errors.has_errors():
            self._display_check_results(errors)

    def _check_implementation_action(self) -> None:
        tab = self._tabs.currentWidget()
        if not isinstance(tab, ProjectTab):
            return

        errors = tab.get_proj().check()
        errors.security_errors = []
        errors.security_warnings = []

        if errors.has_errors():
            self._display_check_results(errors)

    def _new_property_action(self) -> None:
        tab = self._tabs.currentWidget()
        if not isinstance(tab, ProjectTab):
            return

        tab.insertRow(tab.rowCount())

    def _remove_property_action(self) -> None:
        tab = self._tabs.currentWidget()
        if not isinstance(tab, ProjectTab):
            return

        tab.removeRow(tab.currentRow())

    def _check_properties_action(self) -> None:
        tab = self._tabs.currentWidget()
        if not isinstance(tab, ProjectTab):
            return

        errors = tab.get_proj().check()

        if errors.has_errors():
            self._display_check_results(errors)

    def _display_check_results(self, errors: ProjectCheckResult) -> None:
        dlg = QDialog(self)
        dlg.setWindowTitle("Check Results")
        dlg.setMinimumSize(400, 100)
        dlg.setMaximumSize(400, 2160)

        layout = QVBoxLayout()
        dlg.setLayout(layout)

        for key, val in errors.to_dict().items():
            msg = QLabel(dlg)
            msg.setText(key.replace("_", " ").capitalize())
            font = QtGui.QFont()
            font.setBold(True)
            msg.setFont(font)
            dlg.layout().addWidget(msg)

            lst = QListWidget(dlg)
            for item in val:
                lst.addItem(item)
            dlg.layout().addWidget(lst)

        if errors.has_errors():
            dlg.exec()


    def _open_documentation(self) -> None:
        webbrowser.open("https://mavsec.readthedocs.io/en/latest/")

    # Tab Actions
    ################################################################################################
    def _setup_tabs(self) -> None:
        self._prev_tab: None | ProjectTab = None
        self._tabs.currentChanged.connect(self._tab_changed)

    def _tab_changed(self, idx: int) -> None:
        if self._prev_tab is not None:
            self._prev_tab.deactivate()
            self._prev_tab = None

        tab = self._tabs.widget(idx)
        if not isinstance(tab, ProjectTab):
            self._project.clear()
            self._project.setDisabled(True)
            return

        tab.activate()
        self._project.setDisabled(False)
        self._project.set_proj(tab._proj)

        self._prev_tab = tab

    def _setup_toolbar(self) -> None:
        self._new_property.triggered.connect(self._new_property_action)
        self._remove_property.triggered.connect(self._remove_property_action)
        self._check_properties.triggered.connect(self._check_properties_action)

    # GUI Creation
    ################################################################################################
    def _create_tabs(self) -> None:
        self._tabs = QTabWidget(self)
        self._tabs.setObjectName(u"tabs")
        self._tabs.setTabPosition(QTabWidget.TabPosition.South)
        self.setCentralWidget(self._tabs)

    def _create_toolbar(self) -> None:
        self._toolbar = QToolBar("Properties Toolbar", self)
        self.addToolBar(self._toolbar)
        self._toolbar.setIconSize(QtCore.QSize(16, 16))

        fpath = pathlib.Path(__file__).parent

        self._new_property = QtGui.QAction(
                                QtGui.QIcon(str(fpath.joinpath('assets', 'green-plus-hi.png'))),
                                "&New Property",
                                self
                             )
        self._toolbar.addAction(self._new_property)

        self._remove_property = QtGui.QAction(
                                    QtGui.QIcon(str(fpath.joinpath('assets', 'red-minus-hi.png'))),
                                    "&Remove Property",
                                    self
                                )
        self._toolbar.addAction(self._remove_property)

        self._check_properties = QtGui.QAction(
                                   QtGui.QIcon(str(
                                       fpath.joinpath('assets', 'yellow-check-hi.png'))
                                   ),
                                   "&Check Properties",
                                   self
                                 )
        self._toolbar.addAction(self._check_properties)

    # Helper Functions
    ################################################################################################
    def save_tab(self, tab: ProjectTab) -> None:
        if tab.get_proj().info.proj_file is None:
            self.save_tab_as(tab)
        else:
            tab._proj.to_file(tab._proj.info.proj_file)

    def save_tab_as(self, tab: ProjectTab) -> None:
        filename, ok = QFileDialog.getSaveFileName(
            self,
            "Save Project As",
            ".",
            ";;".join(FILE_FILTERS),
        )

        if not ok:
            return

        tab._proj.info.proj_file = filename
        tab.get_proj().to_file(filename)



#####################################################################################
# A tool for the creation of JasperGold SVP principle tcl files.
# Copyright (C) 2024  RISCY-Lib Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#####################################################################################

from __future__ import annotations
from typing import TypeVar, Type

import pathlib

import abc


SchemaT = TypeVar("SchemaT", bound="Schema")


class Schema(abc.ABC):
    @classmethod
    def from_dict(cls: Type[SchemaT], data: dict) -> SchemaT:
        """Get a cls object from a dictionary.

        Args:
            data (dict): The dictionary to convert to a cls object.

        Returns:
            cls: The Schema object.
        """
        return cls(**data)

    @abc.abstractmethod
    def to_dict(self) -> dict:
        """Convert the object to a dictionary.

        Returns:
            dict: The dictionary.
        """
        pass

    @classmethod
    def from_file(cls: Type[SchemaT], path: str | pathlib.Path) -> SchemaT:
        """Get a cls object from a file.

        Args:
            path (str): The path to the file.

        Returns:
            cls: The cls object.
        """
        if isinstance(path, str):
            path = pathlib.Path(path)

        if path.suffix in (".yaml", ".yml"):
            return cls.from_yaml(path)
        elif path.suffix == ".json":
            return cls.from_json(path)
        elif path.suffix == ".toml":
            return cls.from_toml(path)
        else:
            raise ValueError(f"Unsupported file type: {path.suffix}")

    @classmethod
    def from_yaml(cls: Type[SchemaT], path: str | pathlib.Path) -> SchemaT:
        """Get a cls object from a YAML file.

        Args:
            path (str): The path to the YAML file.

        Returns:
            cls: The cls object.
        """
        import yaml

        with open(path, "r") as file:
            data = yaml.safe_load(file)

        return cls.from_dict(data)

    @classmethod
    def from_json(cls: Type[SchemaT], path: str | pathlib.Path) -> SchemaT:
        """Get a cls object from a JSON file.

        Args:
            path (str): The path to the JSON file.

        Returns:
            cls: The cls object.
        """
        import json

        with open(path, "r") as file:
            data = json.load(file)

        return cls.from_dict(data)

    @classmethod
    def from_toml(cls: Type[SchemaT], path: str | pathlib.Path) -> SchemaT:
        """Get a cls object from a TOML file.

        Args:
            path (str): The path to the TOML file.

        Returns:
            cls: The cls object.
        """
        import tomllib

        with open(path, "rb") as file:
            data = tomllib.load(file)

        return cls.from_dict(data)

    def to_yaml(self, path: str | pathlib.Path) -> None:
        """Write the object to a YAML file.

        Args:
            path (str): The path to the YAML file.
        """
        import yaml

        with open(path, "w") as file:
            yaml.dump(self.to_dict(), file)

    def to_json(self, path: str | pathlib.Path) -> None:
        """Write the object to a JSON file.

        Args:
            path (str): The path to the JSON file.
        """
        import json

        with open(path, "w") as file:
            json.dump(self.to_dict(), file)

    def to_toml(self, path: str | pathlib.Path) -> None:
        """Write the object to a TOML file.

        Args:
            path (str): The path to the TOML file.
        """
        import tomli_w

        with open(path, "wb") as file:
            tomli_w.dump(self.to_dict(), file)



#####################################################################################
# A tool for the creation of JasperGold SVP principle tcl files.
# Copyright (C) 2024  RISCY-Lib Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#####################################################################################

from __future__ import annotations

from PySide6 import QtWidgets
import sys
from mavsec._gui import MavSecMainWindow

import re

#####################################################################################
# Version Information
#####################################################################################
from mavsec._info import __version__

version_info = [int(x) if x.isdigit() else x for x in re.split(r"\.|-", __version__)]


#####################################################################################
# GUI Entry Point
#####################################################################################
def gui():
    app = QtWidgets.QApplication(sys.argv)

    window = MavSecMainWindow()
    window.show()

    sys.exit(app.exec())



#####################################################################################
# A tool for the creation of JasperGold SVP principle tcl files.
# Copyright (C) 2024  RISCY-Lib Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#####################################################################################

from __future__ import annotations
from typing import ClassVar, Any, TypeAlias

import enum

from dataclasses import dataclass, field
from mavsec.general import ProjectCheckResult
from mavsec.schema import Schema


class SpecialRtlPaths(enum.StrEnum):
  """Enum for special RTL paths."""
  OUTPUTS = enum.auto()
  INPUTS = enum.auto()


AnyRtlPath = str | SpecialRtlPaths


@dataclass
class PropertyType():

  name: str
  """The name of the property type."""
  description: str
  """A brief description of the property type."""
  meta: dict[str, type | TypeAlias] = field(default_factory=dict)
  """The information needed to generate the property."""

  property_types: ClassVar[dict[str, PropertyType]] = {}

  def __post_init__(self):
    if self.name in self.property_types:
      raise ValueError(f"Property type {self.name} already exists.")

    self.property_types[self.name] = self

  @classmethod
  def get_type(cls, name: str) -> PropertyType:
    """Gets a property type by name."""
    if name not in cls.property_types:
      raise ValueError(f"Property type {name} not found.")
    return cls.property_types[name]

  def check(self, meta: dict, name: str) -> ProjectCheckResult:
    """Check the property type."""
    ret_val = ProjectCheckResult()

    for key, val in self.meta.items():
      if key not in meta:
        ret_val.design_error(f"Property {name} (type {self.name}) missing value for {key}.")
      elif not isinstance(meta[key], val):
        ret_val.design_error(f"Property {name} (type {self.name}) value {key} is not of type {val}.")
      elif isinstance(meta[key], str):
        if meta[key] == '':
          ret_val.design_error(f"Property {name} (type {self.name}) value {key} is empty.")

    return ret_val


SecureKeyProperty = PropertyType(
  "SecureKey",
  "A property that ensures a given key is stored correctly.",
  {"key_loc": AnyRtlPath, "key_size": int, "public_bus": AnyRtlPath}
)

SecureKeyIntegrityProperty = PropertyType(
  "SecureKeyIntegrity",
  "A property that ensures a given key is not overwritten incorrectly.",
  {"key_loc": AnyRtlPath, "key_size": int, "public_bus": AnyRtlPath}
)

SecureKeyGenProperty = PropertyType(
  "SecureKeyGen",
  "A property that ensures a given generated key is stored correctly.",
  {"public_key_loc": AnyRtlPath, "key_size": int, "public_bus": AnyRtlPath}
)

SecureExternalMemoryProperty = PropertyType(
  "SecureExternalMemory",
  "A property that ensures a given external memory is secure."
)

SecureInternalStorageProperty = PropertyType(
  "SecureInternalStorage",
  "A property that ensures a given internal storage is not able to be accessed.",
  {"storage_loc": AnyRtlPath, "storage_size": int, "public_bus": AnyRtlPath}
)

FaultTolerantFSMProperty = PropertyType(
  "FaultTolerantFSM",
  "A property that ensures a given FSM is fault tolerant."
)


@dataclass
class Property(Schema):
  name: str
  """The name of the property."""
  description: str
  """A brief description of the property."""
  meta: dict[str, Any] = field(default_factory=dict)
  """A dictionary of metadata for the property."""
  ptype: PropertyType | str | None = None
  """The type of the property."""
  preconditions: str | list[str] | None = None
  """The preconditions for the property."""

  property_types: ClassVar[list[PropertyType]] = [
    SecureKeyProperty,
    SecureKeyGenProperty,
    SecureExternalMemoryProperty,
    SecureInternalStorageProperty,
    FaultTolerantFSMProperty
  ]
  """Available properties"""

  def __post_init__(self):
    if isinstance(self.ptype, str):
      for ptype in self.property_types:
        if ptype.name == self.ptype:
          self.ptype = ptype
          break
      else:
        raise ValueError(f"Property type {self.ptype} not found.")

  def type_name(self) -> str:
    """Gets the name of the property type."""
    if isinstance(self.ptype, PropertyType):
      return self.ptype.name
    return str(self.ptype)

  def to_svp(self) -> str:
    """Converts the property to an SVP principle."""
    raise NotImplementedError()

  def check(self) -> ProjectCheckResult:
    """Check the property."""
    ret_val = ProjectCheckResult()

    if self.name == '':
      ret_val.security_error("Property name not set.")
    if self.description == '':
      ret_val.security_warning(f"Property ({self.name}) description not set.")
    if self.ptype is None:
      ret_val.security_error(f"Property ({self.name}) type not set.")
      return ret_val

    if isinstance(self.ptype, str):
      self.ptype = PropertyType.get_type(self.ptype)

    return ret_val.merge(self.ptype.check(self.meta, self.name))

  @classmethod
  def ptype_from_str(cls, prop: str) -> PropertyType:
    """Gets a property type from a string."""
    for ptype in cls.property_types:
      if ptype.name == prop:
        return ptype
    raise ValueError(f"Property type {prop} not found.")

  @classmethod
  def from_dict(cls, d: dict) -> Property:
    """Converts the SVP principle to a property."""
    return cls(
      name=d["name"],
      description=d["description"],
      meta=d["meta"],
      ptype=cls.ptype_from_str(d["ptype"])
    )

  @classmethod
  def available_types(cls) -> list[PropertyType]:
    """Returns all the available property types."""
    return cls.property_types

  @classmethod
  def add_type(cls, ptype: PropertyType) -> None:
    """Adds a property type to the available types."""
    cls.property_types.append(ptype)

  def to_dict(self) -> dict:
    """Convert the object to a dictionary."""
    ret_val = {
        "name": self.name,
        "description": self.description,
        "meta": self.meta,
        "ptype": self.type_name()
      }

    return ret_val



#####################################################################################
# A tool for the creation of JasperGold SVP principle tcl files.
# Copyright (C) 2024  RISCY-Lib Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#####################################################################################

from __future__ import annotations


__version__: str = "0.0.1b3"
__author__: str = "Benjamin Davis"



#####################################################################################
# A tool for the creation of JasperGold SVP principle tcl files.
# Copyright (C) 2024  RISCY-Lib Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#####################################################################################

from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class ProjectCheckResult:
  """Class to represent the information when a project is checked."""

  project_errors: list[str] = field(default_factory=list)
  """Errors found in the project setup."""
  project_warnings: list[str] = field(default_factory=list)
  """Warnings found in the project setup."""
  security_errors: list[str] = field(default_factory=list)
  """Errors found during the security handoff check."""
  security_warnings: list[str] = field(default_factory=list)
  """Warnings found during the security handoff check."""
  design_errors: list[str] = field(default_factory=list)
  """Errors found during the design handoff check."""
  design_warnings: list[str] = field(default_factory=list)
  """Warnings found during the design handoff check."""

  def project_error(self, msg: str) -> None:
    """Add an error to the project errors list.

    Args:
        msg (str): The error message.
    """
    self.project_errors.append(msg)

  def project_warning(self, msg: str) -> None:
    """Add a warning to the project warnings list.\

    Args:
        msg (str): The warning message.
    """
    self.project_warnings.append(msg)

  def security_error(self, msg: str) -> None:
    """Add an error to the security errors list.

    Args:
        msg (str): The error message.
    """
    self.security_errors.append(msg)

  def security_warning(self, msg: str) -> None:
    """Add a warning to the security warnings list.

    Args:
        msg (str): The warning message.
    """
    self.security_warnings.append(msg)

  def design_error(self, msg: str) -> None:
    """Add an error to the design errors list.

    Args:
        msg (str): The error message.
    """
    self.design_errors.append(msg)

  def design_warning(self, msg: str) -> None:
    """Add a warning to the design warnings list.

    Args:
        msg (str): The warning message.
    """
    self.design_warnings.append(msg)

  def has_warnings(self) -> bool:
    """Check if there are any warnings.

    Returns:
        bool: True if there are warnings or errors, False otherwise.
    """
    return bool(
      self.project_warnings or self.security_warnings or self.design_warnings or
      self.has_errors()
    )

  def has_errors(self) -> bool:
    """Check if there are any errors.

    Returns:
        bool: True if there are errors, False otherwise.
    """
    return bool(self.project_errors or self.security_errors or self.design_errors)

  def merge(self, other: ProjectCheckResult) -> ProjectCheckResult:
    """Merge two ProjectCheckResult objects.

    Args:
        other (ProjectCheckResult): The other object to merge.

    Returns:
        ProjectCheckResult: The merged object.
    """
    return ProjectCheckResult(
      project_errors=self.project_errors + other.project_errors,
      project_warnings=self.project_warnings + other.project_warnings,
      security_errors=self.security_errors + other.security_errors,
      security_warnings=self.security_warnings + other.security_warnings,
      design_errors=self.design_errors + other.design_errors,
      design_warnings=self.design_warnings + other.design_warnings,
    )

  def merge_into(self, other: ProjectCheckResult) -> None:
    """Merge two ProjectCheckResult objects into this object.

    Args:
        other (ProjectCheckResult): The other object to merge.
    """
    self.project_errors += other.project_errors
    self.project_warnings += other.project_warnings
    self.security_errors += other.security_errors
    self.security_warnings += other.security_warnings
    self.design_errors += other.design_errors
    self.design_warnings += other.design_warnings

  def to_dict(self) -> dict[str, list[str]]:
    """Convert the object to a dictionary.

    Returns:
        dict: The dictionary.
    """
    ret_val = {}

    if len(self.project_errors) != 0:
      ret_val["project_errors"] = self.project_errors
    if len(self.project_warnings) != 0:
      ret_val["project_warnings"] = self.project_warnings
    if len(self.security_errors) != 0:
      ret_val["security_errors"] = self.security_errors
    if len(self.security_warnings) != 0:
      ret_val["security_warnings"] = self.security_warnings
    if len(self.design_errors) != 0:
      ret_val["design_errors"] = self.design_errors
    if len(self.design_warnings) != 0:
      ret_val["design_warnings"] = self.design_warnings

    return ret_val



#####################################################################################
# A tool for the creation of JasperGold SVP principle tcl files.
# Copyright (C) 2024  RISCY-Lib Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#####################################################################################

from __future__ import annotations

from dataclasses import dataclass, field

from mavsec.schema import Schema
from mavsec.properties import Property
from mavsec.general import ProjectCheckResult

import pathlib


@dataclass
class Project(Schema):
    """A project."""

    info: ProjectInfo
    """Information about the project."""
    properties: list[Property] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict) -> Project:
        """Get a Project object from a dictionary.

        Args:
            data (dict): The dictionary to convert to a Project object.

        Returns:
            Project: The Project object.
        """
        return cls(
            info=ProjectInfo.from_dict(data["project"]),
            properties=[Property.from_dict(prop) for prop in data["properties"]],
        )

    def to_dict(self) -> dict:
        """Convert the object to a dictionary.

        Returns:
            dict: The dictionary.
        """
        return {
            "project": self.info.to_dict(),
            "properties": [prop.to_dict() for prop in self.properties],
        }

    def to_file(self, filename: str | pathlib.Path | None = None) -> None:
        """Write the object to a file."""
        if self.info.proj_file is None and filename is None:
            raise ValueError("Project file not set.")

        if filename is not None:
            self.info.proj_file = filename

        if self.info.proj_file is None:
            raise ValueError("Project file not set.")

        filepath = pathlib.Path(self.info.proj_file)

        if filepath.suffix in (".yaml", ".yml"):
            self.to_yaml(filepath)
        elif filepath.suffix == ".json":
            self.to_json(filepath)
        elif filepath.suffix == ".toml":
            self.to_toml(filepath)
        else:
            raise ValueError(f"Unsupported file type: {filepath.suffix}")

    def to_tcl(self, filename: str | pathlib.Path | None = None) -> None:
        """Write the object to a file."""
        raise NotImplementedError("Not implemented yet.")

    def check(self) -> ProjectCheckResult:
        """Check the project for security or implementation issues."""

        issues = ProjectCheckResult()
        issues.merge_into(self.info.check())

        for prop in self.properties:
            issues.merge_into(prop.check())

        return issues


@dataclass
class ProjectInfo(Schema):
    """Information about the project."""

    name: str
    """Name of the project."""
    version: str
    """Version of the project."""
    description: str
    """Description of the project."""
    proj_file: str | pathlib.Path | None = None
    """Path to the project file."""

    def to_dict(self) -> dict:
        """Convert the object to a dictionary.
        proj_file is deliberately not included in the dictionary.

        Returns:
            dict: The dictionary.
        """
        return {
            "name": self.name,
            "version": self.version,
            "description": self.description,
        }

    def check(self) -> ProjectCheckResult:
        """Check the project for issues."""
        issues = ProjectCheckResult()

        if not self.name:
            issues.project_error("Project name is missing.")
        if not self.version:
            issues.project_error("Project version is missing.")

        return issues



#####################################################################################
# A tool for the creation of JasperGold SVP principle tcl files.
# Copyright (C) 2024  RISCY-Lib Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#####################################################################################

from __future__ import annotations

from PySide6.QtWidgets import QStatusBar, QMenuBar, QMenu, QWidget
from PySide6 import QtGui

from mavsec import _info


####################################################################################################
# Main Windows Support Bars
####################################################################################################
class StatusBar(QStatusBar):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setObjectName(u"statusBar")


class MenuFile(QMenu):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setObjectName(u"menuFile")
        self.setTitle("File")

        self.actionNew = QtGui.QAction(parent)
        self.actionNew.setObjectName(u"actionNew")
        self.actionNew.setText(u"New")
        self.actionNew.setShortcut("Ctrl+N")
        self.addAction(self.actionNew)

        self.actionOpen = QtGui.QAction(parent)
        self.actionOpen.setObjectName(u"actionOpen")
        self.actionOpen.setText(u"Open")
        self.actionOpen.setShortcut("Ctrl+O")
        self.addAction(self.actionOpen)

        self.actionSave = QtGui.QAction(parent)
        self.actionSave.setObjectName(u"actionSave")
        self.actionSave.setText(u"Save")
        self.actionSave.setShortcut("Ctrl+S")
        self.addAction(self.actionSave)

        self.actionSaveAs = QtGui.QAction(parent)
        self.actionSaveAs.setObjectName(u"actionSaveAs")
        self.actionSaveAs.setText(u"Save As")
        self.actionSaveAs.setShortcut("Ctrl+Shift+S")
        self.addAction(self.actionSaveAs)

        self.actionSaveAll = QtGui.QAction(parent)
        self.actionSaveAll.setObjectName(u"actionSaveAll")
        self.actionSaveAll.setText(u"Save All")
        self.actionSaveAll.setShortcut("Ctrl+Alt+S")
        self.addAction(self.actionSaveAll)

        self.addSeparator()

        self.actionExport = QtGui.QAction(parent)
        self.actionExport.setObjectName(u"actionExport")
        self.actionExport.setText(u"Export")
        self.actionExport.setShortcut("Ctrl+E")
        self.addAction(self.actionExport)

        self.addSeparator()

        self.actionClose = QtGui.QAction(parent)
        self.actionClose.setObjectName(u"actionClose")
        self.actionClose.setText(u"Close")
        self.actionClose.setShortcut("Ctrl+W")
        self.addAction(self.actionClose)

        self.actionQuit = QtGui.QAction(parent)
        self.actionQuit.setObjectName(u"actionQuit")
        self.actionQuit.setText(u"Quit")
        self.actionQuit.setShortcut("Alt+F4")
        self.addAction(self.actionQuit)


class MenuEdit(QMenu):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setObjectName(u"menuEdit")
        self.setTitle("Edit")

        self.actionNewProperty = QtGui.QAction(parent)
        self.actionNewProperty.setObjectName(u"actionNewProperty")
        self.actionNewProperty.setText(u"New Property")
        self.actionNewProperty.setShortcut("Ctrl+Shift+N")
        self.addAction(self.actionNewProperty)

        self.actionRemoveProperty = QtGui.QAction(parent)
        self.actionRemoveProperty.setObjectName(u"actionRemoveProperty")
        self.actionRemoveProperty.setText(u"Remove Property")
        self.addAction(self.actionRemoveProperty)

        self.addSeparator()

        self.actionCheckSecurity = QtGui.QAction(parent)
        self.actionCheckSecurity.setObjectName(u"actionCheckSecurity")
        self.actionCheckSecurity.setText(u"Check Security Handoff")
        self.addAction(self.actionCheckSecurity)

        self.actionCheckDesign = QtGui.QAction(parent)
        self.actionCheckDesign.setObjectName(u"actionCheckImplementation")
        self.actionCheckDesign.setText(u"Check Design Handoff")
        self.addAction(self.actionCheckDesign)

        self.actionCheckAll = QtGui.QAction(parent)
        self.actionCheckAll.setObjectName(u"actionCheckAll")
        self.actionCheckAll.setText(u"Check All")
        self.actionCheckAll.setShortcut("Ctrl+Shift+C")
        self.addAction(self.actionCheckAll)


class MenuHelp(QMenu):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setObjectName(u"menuHelp")
        self.setTitle("Help")

        self.actionVersion = QtGui.QAction(parent)
        self.actionVersion.setObjectName(u"actionVersion")
        self.actionVersion.setText(f"Version: {_info.__version__}")
        self.actionVersion.setEnabled(False)
        self.addAction(self.actionVersion)

        self.actionDocumentation = QtGui.QAction(parent)
        self.actionDocumentation.setObjectName(u"actionDocumentation")
        self.actionDocumentation.setText(u"Documentation")
        self.addAction(self.actionDocumentation)


class MenuBar(QMenuBar):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setObjectName(u"menuBar")
        self.setGeometry(0, 0, 800, 22)

        self.file = MenuFile(self)
        self.addAction(self.file.menuAction())

        self.edit = MenuEdit(self)
        self.addAction(self.edit.menuAction())

        self.help = MenuHelp(self)
        self.addAction(self.help.menuAction())
